import os


class TerminalColours:
    """
    Usage: https://stackoverflow.com/questions/287871/how-to-print-colored-text-in-python?page=1&tab=votes#tab-top
    """

    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


class ColouredTextFormatter:
    @staticmethod
    def print_coloured_text(text: str, colour: str):
        """
        Prints a text with colour ascii characters.
        """
        return print("".join([colour, text, TerminalColours.ENDC]))


def get_environ_as_type(environment_variable_name: str, default=None, cast_to_type: type = str) -> object:
    """
    Tries to get an environment variable and cast it into a requested type.
    :return: cast_to_type object, or None if failed.
    :raises ValueError: If the value could not be casted into type 'cast_to_type'
    """
    value = os.environ.get(environment_variable_name, default)
    if value is not None:
        try:
            return cast_to_type(value)
        except Exception as e:
            print(e)
            raise ValueError(
                f"Failed to cast environment variable {environment_variable_name} to type {cast_to_type}: the value {value} is not a valid {cast_to_type}"
            )
    return



import inspect
import functools
from typing import Callable


class _ExplicitParamsValidator:
    def __init__(self, function: Callable, validation_type: str = "None"):
        """
        ExplicitParamsValidator
            :param function:
            :param validation_type:
        """
        self.func = function
        self.validation_type = validation_type
        functools.update_wrapper(wrapper=self, wrapped=self.func)

    def __get__(self, obj, type=None):
        return functools.partial(self, obj)

    def __call__(self, *args, **kwargs):
        """
        Caller to create the Wrapper for
            :param args:
            :param kwargs:
            :return:
        """
        if not hasattr(self, "func"):
            self.func = args[0]
            return self

        return self.validate_explicit_params(*args, **kwargs)

    def validate_explicit_params(self, *args, **kwargs):
        """
        validate_explicit_params - Checks if each of the explicit parameters (The ones the user put in the func/method
                                   call) are not None or Empty - Depending on the instantiation
            :param args:  The function arguments
            :param kwargs: The function Keyword arguments
            :return: wrapped function after input params None values validation
        """
        var_names = inspect.getfullargspec(self.func)[0]

        explicit_args_var_names = list(var_names[: len(args)])

        # FOR CLASS METHOD REMOVE THE EXPLICIT DEMAND FOR self PARAMETER
        for params_list in [explicit_args_var_names, list(kwargs.keys())]:
            if "self" in params_list:
                params_list.remove("self")

        # FIRST OF ALL HANDLE ALL OF THE KEYWORD ARGUMENTS
        for kwarg, value in kwargs.items():
            self.__validate_input_param(kwarg, value)

        for i, input_param in enumerate(explicit_args_var_names):
            self.__validate_input_param(input_param, args[i])

        return self.func(*args, **kwargs)

    def __validate_input_param(self, input_param, value):
        """
        __validate_input_param - Validates the input param based on the validation type
                                 in the class constructor
        :param input_param:
        :param value:
        :return:
        """
        if self.validation_type == "NoneOrEmpty":
            if not value:
                raise ValueError("Input param: " + str(input_param) + " is Empty")

        if value is None:
            raise ValueError("Input param: " + str(input_param) + " is None")


# WRAPS THE RETRY DECORATOR CLASS TO ENABLE CALLING WITHOUT PARAMS
def explicit_params_validation(function: Callable = None, validation_type: str = "None"):
    if function is not None:
        return _ExplicitParamsValidator(function=function)
    else:

        def wrapper(function):
            return _ExplicitParamsValidator(function=function, validation_type=validation_type)

        return wrapper



class SingletonMeta(type):
    """
    A Singleton meta class.
    A class that derives from this class will have only 1 instance of that type for the process.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class _SingletonWrapper:
    """
    A singleton wrapper class. Its instances would be created
    for each decorated class.
    """

    def __init__(self, cls):
        self.__wrapped__ = cls
        self._instance = None

    def __call__(self, *args, **kwargs):
        """Returns a single instance of decorated class"""
        if self._instance is None:
            self._instance = self.__wrapped__(*args, **kwargs)
        return self._instance


def singleton(cls):
    """
    A singleton decorator. Returns a wrapper objects. A call on that object
    returns a single instance object of decorated class. Use the __wrapped__
    attribute to access decorated class directly in unit tests
    """
    return _SingletonWrapper(cls)



from abc import ABCMeta
from typing import Any, List, Union

import numpy as np


class AbstractTensorSerializer(ABCMeta):
    """
    Serializes and Deserializes arbitrary data.
    """

    @staticmethod
    def serialize(obj: Union[np.ndarray, List[np.ndarray]]) -> Any:
        """Converts the object to a buffer, which can be later reconstructed using the class's deseralize() method."""
        raise NotImplementedError

    @staticmethod
    def deserialize(obj: Any) -> List[np.ndarray]:
        """
        Deserialized the bytes into a python object.
        """
        raise NotImplementedError
